@@ -17,7 +17,6 @@
 
 #include "senna_in.h"
 
-#include <errno.h>
 #include <string.h>
 #ifdef HAVE_NETDB_H
 #include <netdb.h>
@@ -37,7 +36,7 @@
 
 #include "com.h"
 #include "set.h"
-#include "str.h"
+#include "ctx.h"
 
 #ifndef PF_INET
 #define PF_INET AF_INET
@@ -90,20 +89,19 @@ log_sockerr(const char *msg)
 sen_rc
 sen_com_init(void)
 {
-  sen_rc rc = sen_success;
 #ifdef WIN32
   WSADATA wd;
   if (WSAStartup(MAKEWORD(2, 0), &wd) != 0) {
-    rc = sen_other_error;
+    GERR(sen_external_error, "WSAStartup failed");
   }
 #else /* WIN32 */
 #ifndef USE_MSG_NOSIGNAL
   if (signal(SIGPIPE, SIG_IGN) == SIG_ERR) {
-    rc = sen_other_error;
+    GERR(sen_external_error, "signal");
   }
 #endif /* USE_MSG_NOSIGNAL */
 #endif /* WIN32 */
-  return rc;
+  return sen_gctx.rc;
 }
 
 void
@@ -117,9 +115,10 @@ sen_com_fin(void)
 sen_rc
 sen_com_event_init(sen_com_event *ev, int max_nevents, int data_size)
 {
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
   sen_rc rc = sen_memory_exhausted;
   ev->max_nevents = max_nevents;
-  if ((ev->set = sen_set_open(sizeof(int), data_size, 0))) {
+  if ((ev->set = sen_set_open(sizeof(sen_sock), data_size, 0))) {
 #ifndef USE_SELECT
 #ifdef USE_EPOLL
     if ((ev->events = SEN_MALLOC(sizeof(struct epoll_event) * max_nevents))) {
@@ -128,7 +127,7 @@ sen_com_event_init(sen_com_event *ev, int max_nevents, int data_size)
         goto exit;
       } else {
         LOG_SOCKERR("epoll_create");
-        rc = sen_other_error;
+        rc = sen_external_error;
       }
       SEN_FREE(ev->events);
     }
@@ -152,6 +151,7 @@ exit :
 sen_rc
 sen_com_event_fin(sen_com_event *ev)
 {
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
   if (ev->set) { sen_set_close(ev->set); }
 #ifndef USE_SELECT
   if (ev->events) { SEN_FREE(ev->events); }
@@ -160,32 +160,66 @@ sen_com_event_fin(sen_com_event *ev)
 }
 
 sen_rc
-sen_com_event_add(sen_com_event *ev, int fd, sen_com **com)
+sen_com_event_add(sen_com_event *ev, sen_sock fd, int events, sen_com **com)
 {
   sen_com *c;
-  if (!ev || ev->set->n_entries == ev->max_nevents) { return sen_invalid_argument; }
+  /* todo : expand events */
+  if (!ev || ev->set->n_entries == ev->max_nevents) {
+    if (ev) { SEN_LOG(sen_log_error, "too many connections (%d)", ev->max_nevents); }
+    return sen_invalid_argument;
+  }
 #ifdef USE_EPOLL
   {
     struct epoll_event e;
     memset(&e, 0, sizeof(struct epoll_event));
-    e.events = EPOLLIN;
     e.data.fd = (fd);
+    e.events = (__uint32_t) events;
     if (epoll_ctl(ev->epfd, EPOLL_CTL_ADD, (fd), &e) == -1) {
       LOG_SOCKERR("epoll_ctl");
-      return sen_other_error;
+      return sen_external_error;
     }
   }
 #endif /* USE_EPOLL*/
   if (sen_set_get(ev->set, &fd, (void **)&c)) {
     c->fd = fd;
+    c->events = events;
     if (com) { *com = c; }
     return sen_success;
   }
-  return sen_other_error;
+  return sen_internal_error;
 }
 
 sen_rc
-sen_com_event_del(sen_com_event *ev, int fd)
+sen_com_event_mod(sen_com_event *ev, sen_sock fd, int events, sen_com **com)
+{
+  sen_com *c;
+  if (!ev) { return sen_invalid_argument; }
+  if (sen_set_at(ev->set, &fd, (void **)&c)) {
+    if (c->fd != fd) {
+      SEN_LOG(sen_log_error, "sen_com_event_mod fd unmatch %d != %d", c->fd, fd);
+      return sen_invalid_format;
+    }
+    if (com) { *com = c; }
+    if (c->events != events) {
+#ifdef USE_EPOLL
+      struct epoll_event e;
+      memset(&e, 0, sizeof(struct epoll_event));
+      e.data.fd = (fd);
+      e.events = (__uint32_t) events;
+      if (epoll_ctl(ev->epfd, EPOLL_CTL_MOD, (fd), &e) == -1) {
+        LOG_SOCKERR("epoll_ctl");
+        return sen_external_error;
+      }
+#endif /* USE_EPOLL*/
+      c->events = events;
+    }
+    return sen_success;
+  }
+  return sen_internal_error;
+}
+
+sen_rc
+sen_com_event_del(sen_com_event *ev, sen_sock fd)
 {
   if (!ev) { return sen_invalid_argument; }
 #ifdef USE_EPOLL
@@ -195,7 +229,7 @@ sen_com_event_del(sen_com_event *ev, int fd)
     e.data.fd = fd;
     if (epoll_ctl(ev->epfd, EPOLL_CTL_DEL, fd, &e) == -1) {
       LOG_SOCKERR("epoll_ctl");
-      return sen_other_error;
+      return sen_external_error;
     }
   }
 #endif /* USE_EPOLL*/
@@ -206,24 +240,25 @@ sen_rc
 sen_com_event_poll(sen_com_event *ev, int timeout)
 {
   int nevents;
+  sen_com *com;
 #ifdef USE_SELECT
-  int *pfd, nfds = 0;
+  sen_sock *pfd;
+  int nfds = 0;
   fd_set rfds;
+  fd_set wfds;
   struct timeval tv;
   if (timeout >= 0) {
     tv.tv_sec = timeout / 1000;
     tv.tv_usec = (timeout % 1000) * 1000;
   }
   FD_ZERO(&rfds);
-  {
-    sen_set_cursor *sc = sen_set_cursor_open(ev->set);
-    while (sen_set_cursor_next(sc, (void *)&pfd, NULL)) {
-      FD_SET(*pfd, &rfds);
-      if (*pfd > nfds) { nfds = *pfd; }
-    }
-    sen_set_cursor_close(sc);
-  }
-  nevents = select(nfds + 1, &rfds, NULL, NULL, (timeout >= 0) ? &tv : NULL);
+  FD_ZERO(&wfds);
+  SEN_SET_EACH(ev->set, eh, &pfd, &com, {
+    if ((com->events & SEN_COM_POLLIN)) { FD_SET(*pfd, &rfds); }
+    if ((com->events & SEN_COM_POLLOUT)) { FD_SET(*pfd, &wfds); }
+    if (*pfd > nfds) { nfds = *pfd; }
+  });
+  nevents = select(nfds + 1, &rfds, &wfds, NULL, (timeout >= 0) ? &tv : NULL);
   if (nevents < 0) {
 #ifdef WIN32
     if (WSAGetLastError() == WSAEINTR) { return sen_success; }
@@ -231,19 +266,13 @@ sen_com_event_poll(sen_com_event *ev, int timeout)
     if (errno == EINTR) { return sen_success; }
 #endif /* WIN32 */
     LOG_SOCKERR("select");
-    return sen_other_error;
+    return sen_external_error;
   }
   if (timeout < 0 && !nevents) { SEN_LOG(sen_log_notice, "select returns 0 events"); }
-  {
-    sen_com *com;
-    sen_set_cursor *sc = sen_set_cursor_open(ev->set);
-    while (sen_set_cursor_next(sc, (void *)&pfd, (void *)&com)) {
-      if (FD_ISSET(*pfd, &rfds)) {
-        com->ev_in(ev, com);
-      }
-    }
-    sen_set_cursor_close(sc);
-  }
+  SEN_SET_EACH(ev->set, eh, &pfd, &com, {
+    if (FD_ISSET(*pfd, &rfds)) { com->ev_in(ev, com); }
+    if (FD_ISSET(*pfd, &wfds)) { com->ev_out(ev, com); }
+  });
 #else /* USE_SELECT */
 #ifdef USE_EPOLL
   struct epoll_event *ep;
@@ -251,42 +280,44 @@ sen_com_event_poll(sen_com_event *ev, int timeout)
 #else /* USE_EPOLL */
   int nfd = 0, *pfd;
   struct pollfd *ep = ev->events;
-  sen_set_cursor *sc = sen_set_cursor_open(ev->set);
-  while (sen_set_cursor_next(sc, (void *)&pfd, NULL)) {
+  SEN_SET_EACH(ev->set, eh, &pfd, &com, {
     ep->fd = *pfd;
+    //    ep->events =(short) com->events;
     ep->events = POLLIN;
     ep->revents = 0;
     ep++;
     nfd++;
-  }
-  sen_set_cursor_close(sc);
+  });
   nevents = poll(ev->events, nfd, timeout);
 #endif /* USE_EPOLL */
   if (nevents < 0) {
     if (errno == EINTR) { return sen_success; }
     LOG_SOCKERR("poll");
-    return sen_other_error;
+    return sen_external_error;
   }
   if (timeout < 0 && !nevents) { SEN_LOG(sen_log_notice, "poll returns 0 events"); }
   for (ep = ev->events; nevents; ep++) {
     int efd;
-    sen_com *com;
 #ifdef USE_EPOLL
-    if (!(ep->events & EPOLLIN)) {
-      SEN_LOG(sen_log_error, "unexpected epoll event: %x\n", ep->events);
-      return sen_other_error;
-    }
     efd = ep->data.fd;
-#else /* USE_EPOLL */
-    if (!(ep->revents & POLLIN)) { continue; }
-    efd = ep->fd;
-#endif /* USE_EPOLL */
     nevents--;
     if (!sen_set_at(ev->set, &efd, (void *)&com)) {
       SEN_LOG(sen_log_error, "fd(%d) not found in ev->set", efd);
-      return sen_other_error;
+      return sen_internal_error;
     }
-    com->ev_in(ev, com);
+    if ((ep->events & SEN_COM_POLLIN)) { com->ev_in(ev, com); }
+    if ((ep->events & SEN_COM_POLLOUT)) { com->ev_out(ev, com); }
+#else /* USE_EPOLL */
+    efd = ep->fd;
+    if (!(ep->events & ep->revents)) { continue; }
+    nevents--;
+    if (!sen_set_at(ev->set, &efd, (void *)&com)) {
+      SEN_LOG(sen_log_error, "fd(%d) not found in ev->set", efd);
+      return sen_internal_error;
+    }
+    if ((ep->revents & SEN_COM_POLLIN)) { com->ev_in(ev, com); }
+    if ((ep->revents & SEN_COM_POLLOUT)) { com->ev_out(ev, com); }
+#endif /* USE_EPOLL */
   }
 #endif /* USE_SELECT */
   return sen_success;
@@ -299,17 +330,20 @@ sen_com_sqtp_send(sen_com_sqtp *cs, sen_com_sqtp_header *header, char *body)
 {
   ssize_t ret, whole_size = sizeof(sen_com_sqtp_header) + header->size;
   header->proto = SEN_COM_PROTO_SQTP;
+  if (cs->com.status == sen_com_closing) { header->flags |= SEN_CTX_QUIT; }
+  SEN_LOG(sen_log_info, "send (%d,%x,%d,%02x,%02x,%04x,%08x)", header->size, header->flags, header->proto, header->qtype, header->level, header->status, header->info);
+
   if (header->size) {
 #ifdef WIN32
     ssize_t reth;
     if ((reth = send(cs->com.fd, header, sizeof(sen_com_sqtp_header), 0)) == -1) {
       LOG_SOCKERR("send size");
-      cs->rc = sen_other_error;
+      cs->rc = sen_external_error;
       goto exit;
     }
     if ((ret = send(cs->com.fd, body, header->size, 0)) == -1) {
       LOG_SOCKERR("send body");
-      cs->rc = sen_other_error;
+      cs->rc = sen_external_error;
       goto exit;
     }
     ret += reth;
@@ -330,7 +364,7 @@ sen_com_sqtp_send(sen_com_sqtp *cs, sen_com_sqtp_header *header, char *body)
     while ((ret = sendmsg(cs->com.fd, &msg, MSG_NOSIGNAL)) == -1) {
       LOG_SOCKERR("sendmsg");
       if (errno == EAGAIN || errno == EINTR) { continue; }
-      cs->rc = sen_other_error;
+      cs->rc = sen_external_error;
       goto exit;
     }
 #endif /* WIN32 */
@@ -344,13 +378,13 @@ sen_com_sqtp_send(sen_com_sqtp *cs, sen_com_sqtp_header *header, char *body)
       LOG_SOCKERR("send");
       if (errno == EAGAIN || errno == EINTR) { continue; }
 #endif /* WIN32 */
-      cs->rc = sen_other_error;
+      cs->rc = sen_external_error;
       goto exit;
     }
   }
   if (ret != whole_size) {
     SEN_LOG(sen_log_error, "sendmsg: %d < %d", ret, whole_size);
-    cs->rc = sen_other_error;
+    cs->rc = sen_external_error;
     goto exit;
   }
   cs->rc = sen_success;
@@ -363,6 +397,7 @@ sen_com_sqtp_recv(sen_com_sqtp *cs, sen_rbuf *buf,
                   unsigned int *status, unsigned int *info)
 {
   ssize_t ret;
+  sen_com_sqtp_header *header;
   size_t rest = sizeof(sen_com_sqtp_header);
   if (SEN_RBUF_WSIZE(buf) < rest) {
     if ((cs->rc = sen_rbuf_reinit(buf, rest))) {
@@ -386,21 +421,25 @@ sen_com_sqtp_recv(sen_com_sqtp *cs, sen_rbuf *buf,
         *info = errno;
 #endif /* WIN32 */
       }
-      cs->rc = sen_other_error;
+      cs->rc = sen_external_error;
       *status = sen_com_erecv_head;
       goto exit;
     }
     rest -= ret, buf->curr += ret;
   } while (rest);
+  header = SEN_COM_SQTP_MSG_HEADER(buf);
+  SEN_LOG(sen_log_info, "recv (%d,%x,%d,%02x,%02x,%04x,%08x)", header->size, header->flags, header->proto, header->qtype, header->level, header->status, header->info);
+  *status = header->status;
+  *info = header->info;
   {
-    sen_com_sqtp_header *header = SEN_COM_SQTP_MSG_HEADER(buf);
+    uint16_t proto = header->proto;
     size_t value_size = header->size;
     size_t whole_size = sizeof(sen_com_sqtp_header) + value_size;
-    if (header->proto != SEN_COM_PROTO_SQTP) {
-      SEN_LOG(sen_log_error, "illegal header: %d", header->proto);
+    if (proto != SEN_COM_PROTO_SQTP) {
+      SEN_LOG(sen_log_error, "illegal header: %d", proto);
       cs->rc = sen_invalid_format;
       *status = sen_com_eproto;
-      *info = header->proto;
+      *info = proto;
       goto exit;
     }
     if (SEN_RBUF_WSIZE(buf) < whole_size) {
@@ -423,15 +462,13 @@ sen_com_sqtp_recv(sen_com_sqtp *cs, sen_rbuf *buf,
           *info = errno;
 #endif /* WIN32 */
         }
-        cs->rc = sen_other_error;
+        cs->rc = sen_external_error;
         *status = sen_com_erecv_body;
         goto exit;
       }
       rest -= ret, buf->curr += ret;
     }
     *buf->curr = '\0';
-    *status = header->status;
-    *info = header->info;
   }
   cs->rc = sen_success;
 exit :
@@ -441,18 +478,21 @@ exit :
 sen_com_sqtp *
 sen_com_sqtp_copen(sen_com_event *ev, const char *dest, int port)
 {
-  int fd;
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
+  sen_sock fd;
   sen_com_sqtp *cs = NULL;
   struct hostent *he;
   struct sockaddr_in addr;
   if (!(he = gethostbyname(dest))) {
     LOG_SOCKERR("gethostbyname");
+    goto exit;
   }
   addr.sin_family = AF_INET;
   memcpy(&addr.sin_addr, he->h_addr, he->h_length);
   addr.sin_port = htons(port);
   if ((fd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
     LOG_SOCKERR("socket");
+    goto exit;
   }
   {
     int v = 1;
@@ -475,14 +515,14 @@ sen_com_sqtp_copen(sen_com_event *ev, const char *dest, int port)
     goto exit;
   }
   if (ev) {
-    if (sen_com_event_add(ev, fd, (sen_com **)&cs)) { goto exit; }
+    if (sen_com_event_add(ev, fd, SEN_COM_POLLIN, (sen_com **)&cs)) { goto exit; }
   } else {
     if (!(cs = SEN_CALLOC(sizeof(sen_com_sqtp)))) { goto exit; }
     cs->com.fd = fd;
   }
   sen_rbuf_init(&cs->msg, 0);
 exit :
-  if (!cs) { close(fd); }
+  if (!cs) { sen_sock_close(fd); }
   return cs;
 }
 
@@ -491,9 +531,17 @@ sen_com_sqtp_receiver(sen_com_event *ev, sen_com *c)
 {
   unsigned int status, info;
   sen_com_sqtp *cs = (sen_com_sqtp *)c;
-  if (sen_com_sqtp_recv(cs, &cs->msg, &status, &info)) {
-    // todo : must be handled
+  if (cs->com.status == sen_com_closing) {
+    sen_com_sqtp_close(ev, cs);
+    return;
   }
+  if (cs->com.status != sen_com_idle) {
+    SEN_LOG(sen_log_info, "waiting to be idle.. (%d) %d", c->fd, ev->set->n_entries);
+    usleep(1000);
+    return;
+  }
+  sen_com_sqtp_recv(cs, &cs->msg, &status, &info);
+  cs->com.status = sen_com_doing;
   cs->msg_in(ev, c);
 }
 
@@ -501,13 +549,13 @@ static void
 sen_com_sqtp_acceptor(sen_com_event *ev, sen_com *c)
 {
   sen_com_sqtp *cs = (sen_com_sqtp *)c, *ncs;
-  int fd = accept(cs->com.fd, NULL, NULL);
+  sen_sock fd = accept(cs->com.fd, NULL, NULL);
   if (fd == -1) {
     LOG_SOCKERR("accept");
     return;
   }
-  if (sen_com_event_add(ev, fd, (sen_com **)&ncs)) {
-    close(fd);
+  if (sen_com_event_add(ev, fd, SEN_COM_POLLIN, (sen_com **)&ncs)) {
+    sen_sock_close(fd);
     return;
   }
   ncs->com.ev_in = sen_com_sqtp_receiver;
@@ -515,12 +563,13 @@ sen_com_sqtp_acceptor(sen_com_event *ev, sen_com *c)
   ncs->msg_in = cs->msg_in;
 }
 
-#define LISTEN_BACKLOG 50
+#define LISTEN_BACKLOG 0x1000
 
 sen_com_sqtp *
 sen_com_sqtp_sopen(sen_com_event *ev, int port, sen_com_callback *func)
 {
-  int lfd;
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
+  sen_sock lfd;
   sen_com_sqtp *cs = NULL;
   struct sockaddr_in addr;
   memset(&addr, 0, sizeof(struct sockaddr_in));
@@ -562,7 +611,7 @@ sen_com_sqtp_sopen(sen_com_event *ev, int port, sen_com_callback *func)
     goto exit;
   }
   if (ev) {
-    if (sen_com_event_add(ev, lfd, (sen_com **)&cs)) { goto exit; }
+    if (sen_com_event_add(ev, lfd, SEN_COM_POLLIN, (sen_com **)&cs)) { goto exit; }
   } else {
     if (!(cs = SEN_MALLOC(sizeof(sen_com_sqtp)))) { goto exit; }
     cs->com.fd = lfd;
@@ -572,7 +621,7 @@ exit :
     cs->com.ev_in = sen_com_sqtp_acceptor;
     cs->msg_in = func;
   } else {
-    close(lfd);
+    sen_sock_close(lfd);
   }
   return cs;
 }
@@ -580,12 +629,18 @@ exit :
 sen_rc
 sen_com_sqtp_close(sen_com_event *ev, sen_com_sqtp *cs)
 {
-  int fd = cs->com.fd;
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
+  sen_sock fd = cs->com.fd;
   sen_rbuf_fin(&cs->msg);
   if (ev) {
     sen_com_event_del(ev, fd);
   } else {
     SEN_FREE(cs);
   }
-  return close(fd) ? sen_other_error : sen_success;
+  if (shutdown(fd, SHUT_RDWR) == -1) { /* LOG_SOCKERR("shutdown"); */ }
+  if (sen_sock_close(fd) == -1) {
+    LOG_SOCKERR("close");
+    return sen_external_error;
+  }
+  return sen_success;
 }
