@@ -499,6 +499,13 @@ exit :
   return rc;
 }
 
+static void
+token_info_clear_offset(token_info **tis, uint32_t n)
+{
+  token_info **tie;
+  for (tie = tis + n; tis < tie; tis++) { (*tis)->offset = 0; }
+}
+
 /* sen_index */
 
 inline static int
@@ -1414,11 +1421,12 @@ sen_records_rewind(sen_records *r)
   if (!r) { return sen_invalid_argument; }
   if (r->sorted) {
     r->curr_rec = NULL;
+    return sen_success;
   } else {
     sen_records_cursor_clear(r);
     r->cursor = sen_set_cursor_open(r->records);
+    return r->cursor ? sen_success : sen_internal_error;
   }
-  return r->cursor ? sen_success : sen_internal_error;
 }
 
 int
@@ -2287,9 +2295,17 @@ sen_index_select(sen_index *i, const char *string, unsigned int string_len,
   }
   r->keys = i->keys;
   if (token_info_build(i, string, string_len, tis, &n, mode) || !n) { goto exit; }
-  if (mode == sen_sel_near) {
+  switch (mode) {
+  case sen_sel_near2 :
+    token_info_clear_offset(tis, n);
+    mode = sen_sel_near;
+    /* fallthru */
+  case sen_sel_near :
     if (!(bt = bt_open(n))) { rc = sen_memory_exhausted; goto exit; }
     max_interval = optarg->max_interval;
+    break;
+  default :
+    break;
   }
   qsort(tis, n, sizeof(token_info *), token_compare);
   tie = tis + n;
