@@ -64,7 +64,7 @@ sen_inv *sen_inv_create(const char *path, sen_sym *lexicon,
 sen_inv *sen_inv_open(const char *path, sen_sym *lexicon);
 sen_rc sen_inv_close(sen_inv *inv);
 sen_rc sen_inv_remove(const char *path);
-sen_rc sen_inv_info(sen_inv *inv, off_t *seg_size, off_t *chunk_size);
+sen_rc sen_inv_info(sen_inv *inv, uint64_t *seg_size, uint64_t *chunk_size);
 sen_rc sen_inv_update(sen_inv *inv, uint32_t key, sen_inv_updspec *u, sen_set *h,
                       int hint);
 sen_rc sen_inv_delete(sen_inv *inv, uint32_t key, sen_inv_updspec *u, sen_set *h);
