@@ -188,18 +188,18 @@ sen_inv_seg_expire08(sen_inv *inv)
         sen_io_mapinfo *info = &inv->seg->maps[pseg];
         uint32_t *pnref = &inv->seg->nrefs[pseg];
         if (info->map && !*pnref) {
-          sen_io_seg_expire(inv->seg, pseg);
+          sen_io_seg_expire(inv->seg, pseg, 100);
           if (inv->seg->nmaps <= expire_threshold) { return; }
         }
       }
     }
     for (seg = inv->amax; seg; seg--) {
       uint16_t pseg = inv->ainfo[seg];
-      if (pseg != SEG_NOT_ASSIGNED) {      
+      if (pseg != SEG_NOT_ASSIGNED) {
         sen_io_mapinfo *info = &inv->seg->maps[pseg];
         uint32_t *pnref = &inv->seg->nrefs[pseg];
         if (info->map && !*pnref) {
-          sen_io_seg_expire(inv->seg, pseg);
+          sen_io_seg_expire(inv->seg, pseg, 100);
           if (inv->seg->nmaps <= expire_threshold) { return; }
         }
       }
@@ -653,7 +653,7 @@ encode_rec(sen_inv_updspec *u, unsigned int *size, int deletep)
   uint8_t *br, *p;
   struct _sen_inv_pos *pp;
   uint32_t lpos, tf = deletep ? 0 : u->tf;
-  if (!(br = SEN_MALLOC((u->tf + 4) * 5))) {
+  if (!(br = SEN_GMALLOC((u->tf + 4) * 5))) {
     return NULL;
   }
   p = br;
@@ -733,6 +733,7 @@ sym_delete(sen_inv *inv, uint32_t tid, sen_set *h)
 inline static sen_rc
 buffer_flush(sen_inv *inv, uint32_t seg, sen_set *h)
 {
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
   buffer *sb, *db = NULL;
   sen_rc rc = sen_success;
   sen_io_win sw, dw;
@@ -765,7 +766,7 @@ buffer_flush(sen_inv *inv, uint32_t seg, sen_set *h)
   }
   //  sen_log("db=%p ds=%d sb=%p seg=%d", db, ds, sb, seg);
   if ((scn = sb->header.chunk) != CHUNK_NOT_ASSIGNED) {
-    sc = sen_io_win_map(inv->chunk, &sw, scn, 0, sb->header.chunk_size, sen_io_rdwr);
+    sc = sen_io_win_map(inv->chunk, ctx, &sw, scn, 0, sb->header.chunk_size, sen_io_rdwr);
     if (!sc) {
       SEN_LOG(sen_log_alert, "io_win_map(%d, %d) failed!!", scn, sb->header.chunk_size);
       buffer_close(inv, seg * SEN_INV_SEGMENT_SIZE);
@@ -775,7 +776,7 @@ buffer_flush(sen_inv *inv, uint32_t seg, sen_set *h)
     }
   }
   // dc = sen_io_win_map(inv->chunk, &dw, dcn, 0, max_dest_chunk_size, sen_io_wronly);
-  dc = sen_io_win_map(inv->chunk, &dw, dcn, 0, max_dest_chunk_size, sen_io_rdwr);
+  dc = sen_io_win_map(inv->chunk, ctx, &dw, dcn, 0, max_dest_chunk_size, sen_io_rdwr);
   if (!dc) { 
     SEN_LOG(sen_log_alert, "io_win_map(%d, %d) failed!!", dcn, max_dest_chunk_size);
     buffer_close(inv, seg * SEN_INV_SEGMENT_SIZE);
@@ -1055,7 +1056,7 @@ sen_inv_create08(const char *path, sen_sym *lexicon, uint32_t initial_n_segments
   memcpy(header->idstr, SEN_INV_IDSTR, 16);
   for (i = 0; i < SEN_INV_MAX_SEGMENT; i++) { header->segments[i] = 0; }
   header->initial_n_segments = initial_n_segments;
-  if (!(inv = SEN_MALLOC(sizeof(sen_inv)))) {
+  if (!(inv = SEN_GMALLOC(sizeof(sen_inv)))) {
     sen_io_close(seg);
     sen_io_close(chunk);
     return NULL;
@@ -1088,7 +1089,7 @@ sen_inv_open08(const char *path, sen_sym *lexicon)
     return NULL;
   }
   header = sen_io_header(seg);
-  if (!(inv = SEN_MALLOC(sizeof(sen_inv)))) {
+  if (!(inv = SEN_GMALLOC(sizeof(sen_inv)))) {
     sen_io_close(seg);
     sen_io_close(chunk);
     return NULL;
@@ -1172,7 +1173,7 @@ sen_inv_update08(sen_inv *inv, uint32_t key, sen_inv_updspec *u, sen_set *h, int
       goto exit;
     }
           if ((r = buffer_new(inv, size + size2, &pos, &bt, &br, &b, hint))) {
-            SEN_FREE(bs2);
+            SEN_GFREE(bs2);
             goto exit;
           }
           bt->tid = key;
@@ -1181,12 +1182,12 @@ sen_inv_update08(sen_inv *inv, uint32_t key, sen_inv_updspec *u, sen_set *h, int
           bt->size_in_buffer = 0;
           bt->pos_in_buffer = 0;
           if ((r = buffer_put(b, bt, br, bs2, &u2, size2))) {
-            SEN_FREE(bs2);
+            SEN_GFREE(bs2);
             buffer_close(inv, pos);
             goto exit;
           }
           br = (buffer_rec *)(((byte *)br) + size2);
-          SEN_FREE(bs2);
+          SEN_GFREE(bs2);
         }
       }
     }
@@ -1215,7 +1216,7 @@ sen_inv_update08(sen_inv *inv, uint32_t key, sen_inv_updspec *u, sen_set *h, int
   }
 exit :
   array_unref(inv, key);
-  if (bs) { SEN_FREE(bs); }
+  if (bs) { SEN_GFREE(bs); }
   return r;
 }
 
@@ -1275,7 +1276,7 @@ sen_inv_delete08(sen_inv *inv, uint32_t key, sen_inv_updspec *u, sen_set *h)
   }
 exit :
   array_unref(inv, key);
-  if (bs) { SEN_FREE(bs); }
+  if (bs) { SEN_GFREE(bs); }
   return r;
 }
 
@@ -1287,11 +1288,12 @@ exit :
 sen_inv_cursor *
 sen_inv_cursor_open08(sen_inv *inv, uint32_t key)
 {
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
   sen_inv_cursor08 *c  = NULL;
   uint32_t pos, *a = array_at(inv, key);
   if (!a) { return NULL; }
   if (!(pos = *a)) { goto exit; }
-  if (!(c = SEN_MALLOC(sizeof(sen_inv_cursor08)))) { goto exit; }
+  if (!(c = SEN_GMALLOC(sizeof(sen_inv_cursor08)))) { goto exit; }
   memset(c, 0, sizeof(sen_inv_cursor08));
   c->inv = inv;
   if (pos & 1) {
@@ -1307,16 +1309,16 @@ sen_inv_cursor_open08(sen_inv *inv, uint32_t key)
     c->pb.rid = 0; c->pb.sid = 0; /* for check */
     c->buffer_pos = pos;
     if (buffer_open(inv, pos, &bt, &c->buf)) {
-      SEN_FREE(c);
+      SEN_GFREE(c);
       c = NULL;
       goto exit;
     }
     if (bt->size_in_chunk && (chunk = c->buf->header.chunk) != CHUNK_NOT_ASSIGNED) {
-      c->cp = sen_io_win_map(inv->chunk, &c->iw,
+      c->cp = sen_io_win_map(inv->chunk, ctx, &c->iw,
                              chunk, bt->pos_in_chunk, bt->size_in_chunk, sen_io_rdonly);
       if (!c->cp) {
         buffer_close(inv, pos);
-        SEN_FREE(c);
+        SEN_GFREE(c);
         c = NULL;
         goto exit;
       }
@@ -1468,7 +1470,7 @@ sen_inv_cursor_close08(sen_inv_cursor *c1)
   if (!c) { return sen_invalid_argument; }
   if (c->cp) { sen_io_win_unmap(&c->iw); }
   if (c->buf) { buffer_close(c->inv, c->buffer_pos); }
-  SEN_FREE(c);
+  SEN_GFREE(c);
   return sen_success;
 }
 
