@@ -112,6 +112,29 @@ void * sen_io_seg_ref(sen_io *io, uint32_t segno);
 sen_rc sen_io_seg_unref(sen_io *io, uint32_t segno);
 sen_rc sen_io_seg_expire(sen_io *io, uint32_t segno, uint32_t nretry);
 
+typedef struct _sen_io_ja_einfo sen_io_ja_einfo;
+typedef struct _sen_io_ja_ehead sen_io_ja_ehead;
+
+struct _sen_io_ja_einfo {
+  uint32_t pos;
+  uint32_t size;
+};
+
+struct _sen_io_ja_ehead {
+  uint32_t size;
+  uint32_t key;
+};
+
+sen_rc sen_io_read_ja(sen_io *io, sen_ctx *ctx, sen_io_ja_einfo *einfo, uint32_t epos,
+                      uint32_t key, uint32_t segment, uint32_t offset,
+                      void **value, uint32_t *value_len);
+sen_rc sen_io_write_ja(sen_io *io, sen_ctx *ctx,
+                       uint32_t key, uint32_t segment, uint32_t offset,
+                       void *value, uint32_t value_len);
+
+sen_rc sen_io_write_ja_ehead(sen_io *io, sen_ctx *ctx, uint32_t key,
+                             uint32_t segment, uint32_t offset, uint32_t value_len);
+
 #define SEN_IO_MAX_REF 0x80000000
 #define SEN_IO_MAX_RETRY 0x10000
 
