@@ -204,6 +204,13 @@ get_op(sen_query *q, sen_sel_operator op, int weight)
     if (start == end) { option = DEFAULT_MAX_INTERVAL; }
     q->cur = end;
     break;
+  case 'n' :
+    mode = sen_sel_near2;
+    start = ++end;
+    option = sen_atoi(start, q->str_end, (const char **)&end);
+    if (start == end) { option = DEFAULT_MAX_INTERVAL; }
+    q->cur = end;
+    break;
   case 'T' :
     mode = sen_sel_term_extract;
     start = ++end;
@@ -482,7 +489,7 @@ exec_query(sen_index *i, sen_query *q, cell *c, sen_records *r, sen_sel_operator
   int n = sen_records_nhits(r);
   sen_sel_operator op0 = sen_sel_or, *opp = &op0, op1 = q->default_op;
   if (!n && op != sen_sel_or) { return; }
-  s = n ? sen_records_open(sen_rec_document, sen_rec_none, 0) : r;
+  s = n ? sen_records_open(r->record_unit, r->subrec_unit, 0) : r;
   while (c != NIL) {
     POP(e, c);
     switch (e->type) {
@@ -616,7 +623,7 @@ scan_keyword(snip_cond *sc, sen_nstr *str, sen_id section,
   int tf;
   int w = 1;
   for (tf = 0; ; tf++) {
-    sen_bm_tunedbm(sc, str);
+    sen_bm_tunedbm(sc, str, 0);
     if (sc->stopflag == SNIPCOND_STOP) { break; }
   }
   if (optarg->vector_size) {
@@ -636,8 +643,9 @@ scan_keyword(snip_cond *sc, sen_nstr *str, sen_id section,
     break;
   case sen_sel_and :
     if (tf) {
-      *found &= 1;
       *score += w * tf;
+    } else {
+      *found = 0;
     }
     break;
   case sen_sel_but :
