@@ -174,7 +174,7 @@ sen_ql_at(sen_ctx *ctx, const char *key)
 void
 sen_ql_def_native_func(sen_ctx *ctx, const char *name, sen_ql_native_func *func)
 {
-  sen_obj *o = sen_ql_mk_symbol(ctx, name);
+  sen_obj *o = INTERN(name);
   if (o != F) {
     o->type = sen_ql_void;
     o->flags |= SEN_OBJ_NATIVE;
@@ -365,6 +365,7 @@ enum {
   OP_PUT,
   OP_GET,
   OP_QUIT,
+  OP_SDOWN,
   OP_RDSEXPR,
   OP_RDLIST,
   OP_RDDOT,
@@ -501,11 +502,11 @@ skipline(sen_ctx *ctx)
 #define CONT_DUMP(p)     CDR(p)
 #define PROMISEP(p)      ((p)->flags & SEN_OBJ_PROMISE)
 #define SETPROMISE(p)    (p)->flags |= SEN_OBJ_PROMISE
-#define LAMBDA           (sen_ql_mk_symbol(ctx, "lambda"))
-#define QUOTE            (sen_ql_mk_symbol(ctx, "quote"))
-#define QQUOTE           (sen_ql_mk_symbol(ctx, "quasiquote"))
-#define UNQUOTE          (sen_ql_mk_symbol(ctx, "unquote"))
-#define UNQUOTESP        (sen_ql_mk_symbol(ctx, "unquote-splicing"))
+#define LAMBDA           (INTERN("lambda"))
+#define QUOTE            (INTERN("quote"))
+#define QQUOTE           (INTERN("quasiquote"))
+#define UNQUOTE          (INTERN("unquote"))
+#define UNQUOTESP        (INTERN("unquote-splicing"))
 
 /* get new cell.  parameter a, b is marked by gc. */
 #define GET_CELL(ctx,a,b,o) SEN_OBJ_NEW(ctx, o)
@@ -551,7 +552,7 @@ sen_ql_mk_symbol2(sen_ctx *ctx, const char *q, unsigned int len, int kwdp)
   if (kwdp) { *p++ = ':'; }
   memcpy(p, q, len);
   p[len] = '\0';
-  return sen_ql_mk_symbol(ctx, buf);
+  return INTERN(buf);
 }
 
 inline static cell *
@@ -1717,6 +1718,11 @@ opexe(sen_ctx *ctx)
       s_return(ctx, NIL);
     }
 
+  case OP_SDOWN:   /* shutdown */
+    SEN_LOG(sen_log_notice, "shutting down..");
+    sen_gctx.stat = SEN_CTX_QUIT;
+    s_goto(ctx, OP_QUIT);
+
   case OP_RDSEXPR:
     {
       char tok, *str;
@@ -2016,48 +2022,48 @@ sen_ql_eval(sen_ctx *ctx, sen_obj *code, sen_obj *objs)
   }\
 } while (0)
 
-#define do_compare(x,y,op) do {\
+#define do_compare(x,y,r,op) do {\
   switch (x->type) {\
   case sen_ql_int :\
     switch (y->type) {\
     case sen_ql_int :\
-      s_retbool(IVALUE(x) op IVALUE(y));\
+      r = (IVALUE(x) op IVALUE(y));\
       break;\
     case sen_ql_float :\
-      s_retbool(IVALUE(x) op FVALUE(y));\
+      r = (IVALUE(x) op FVALUE(y));\
       break;\
     default :\
       if (sen_obj2int(ctx, y)) { QLERR("can't convert into numeric value"); }\
-      s_retbool(IVALUE(x) op IVALUE(y));\
+      r = (IVALUE(x) op IVALUE(y));\
     }\
     break;\
   case sen_ql_float :\
     switch (y->type) {\
     case sen_ql_int :\
-      s_retbool(FVALUE(x) op IVALUE(y));\
+      r = (FVALUE(x) op IVALUE(y));\
       break;\
     case sen_ql_float :\
-      s_retbool(FVALUE(x) op FVALUE(y));\
+      r = (FVALUE(x) op FVALUE(y));\
       break;\
     default :\
       if (sen_obj2int(ctx, y)) { QLERR("can't convert into numeric value"); }\
-      s_retbool(FVALUE(x) op IVALUE(y));\
+      r = (FVALUE(x) op IVALUE(y));\
     }\
     break;\
   case sen_ql_bulk :\
     if (y->type == sen_ql_bulk) {\
-      int r;\
+      int r_;\
       uint32_t la = x->u.b.size, lb = y->u.b.size;\
       if (la > lb) {\
-        if (!(r = memcmp(x->u.b.value, y->u.b.value, lb))) {\
-          r = 1;\
+        if (!(r_ = memcmp(x->u.b.value, y->u.b.value, lb))) {\
+          r_ = 1;\
         }\
       } else {\
-        if (!(r = memcmp(x->u.b.value, y->u.b.value, la))) {\
-          r = la == lb ? 0 : -1;\
+        if (!(r_ = memcmp(x->u.b.value, y->u.b.value, la))) {\
+          r_ = la == lb ? 0 : -1;\
         }\
       }\
-      s_retbool(r op 0);\
+      r = (r_ op 0);\
     } else {\
       QLERR("can't compare");\
     }\
@@ -2065,14 +2071,15 @@ sen_ql_eval(sen_ctx *ctx, sen_obj *code, sen_obj *objs)
   case sen_ql_time :\
     if (y->type == sen_ql_time) {\
       if (x->u.tv.tv_sec != y->u.tv.tv_sec) {\
-        s_retbool(x->u.tv.tv_sec op y->u.tv.tv_sec);\
+        r = (x->u.tv.tv_sec op y->u.tv.tv_sec);\
+      } else {\
+        r = (x->u.tv.tv_usec op y->u.tv.tv_usec);\
       }\
-      s_retbool(x->u.tv.tv_usec op y->u.tv.tv_usec);\
     } else {\
       QLERR("can't compare");\
     }\
   default :\
-    s_retbool(memcmp(&x->u.tv, &y->u.tv, sizeof(sen_timeval)) op 0);\
+    r = (memcmp(&x->u.tv, &y->u.tv, sizeof(sen_timeval)) op 0);\
   }\
 } while (0)
 
@@ -2123,7 +2130,7 @@ static sen_obj *
 nf_add(sen_ctx *ctx, sen_obj *args, sen_ql_co *co)
 {
   register cell *x, *v;
-  if (!PAIRP(args)) { QLERR("list expected"); }
+  if (!PAIRP(args)) { QLERR("list required"); }
   switch (CAR(args)->type) {
   case sen_ql_bulk :
     {
@@ -2329,35 +2336,129 @@ nf_negp(sen_ctx *ctx, sen_obj *args, sen_ql_co *co)
 static sen_obj *
 nf_neq(sen_ctx *ctx, sen_obj *args, sen_ql_co *co)
 {
-  s_retbool(IVALUE(CAR(args)) == IVALUE(CADR(args)));
+  int r = 1;
+  register cell *x, *y;
+  POP(x, args);
+  if (!PAIRP(args)) { QLERR("Few arguments"); }
+  do {
+    POP(y, args);
+    switch (x->type) {
+    case sen_ql_int :
+      switch (y->type) {
+      case sen_ql_int :
+        r = (IVALUE(x) == IVALUE(y));
+        break;
+      case sen_ql_float :
+        r = (IVALUE(x) <= FVALUE(y) && IVALUE(x) >= FVALUE(y));
+        break;
+      default :
+        if (sen_obj2int(ctx, y)) { QLERR("can't convert into numeric value"); }
+        r = (IVALUE(x) == IVALUE(y));
+      }
+      break;
+    case sen_ql_float :
+      switch (y->type) {
+      case sen_ql_int :
+        r = (FVALUE(x) <= IVALUE(y) && FVALUE(x) >= IVALUE(y));
+        break;
+      case sen_ql_float :
+        r = (FVALUE(x) <= FVALUE(y) && FVALUE(x) >= FVALUE(y));
+        break;
+      default :
+        if (sen_obj2int(ctx, y)) { QLERR("can't convert into numeric value"); }
+        r = (FVALUE(x) <= IVALUE(y) && FVALUE(x) >= IVALUE(y));
+      }
+      break;
+    case sen_ql_bulk :
+      if (y->type == sen_ql_bulk) {
+        int r_;
+        uint32_t la = x->u.b.size, lb = y->u.b.size;
+        if (la > lb) {
+          if (!(r_ = memcmp(x->u.b.value, y->u.b.value, lb))) {
+            r_ = 1;
+          }
+        } else {
+          if (!(r_ = memcmp(x->u.b.value, y->u.b.value, la))) {
+            r_ = la == lb ? 0 : -1;
+          }
+        }
+        r = (r_ == 0);
+      } else {
+        QLERR("can't compare");
+      }
+      break;
+    case sen_ql_time :
+      if (y->type == sen_ql_time) {
+        if (x->u.tv.tv_sec != y->u.tv.tv_sec) {
+          r = (x->u.tv.tv_sec == y->u.tv.tv_sec);
+        } else {
+          r = (x->u.tv.tv_usec == y->u.tv.tv_usec);
+        }
+      } else {
+        QLERR("can't compare");
+      }
+    default :
+      r = (memcmp(&x->u.tv, &y->u.tv, sizeof(sen_timeval)) == 0);
+    }
+    x = y;
+  } while (PAIRP(args) && r);
+  return r ? T : F;
 }
 static sen_obj *
 nf_less(sen_ctx *ctx, sen_obj *args, sen_ql_co *co)
 {
-  register cell *x = CAR(args);
-  register cell *y = CADR(args);
-  do_compare(x, y, <);
+  int r = 1;
+  register cell *x, *y;
+  POP(x, args);
+  if (!PAIRP(args)) { QLERR("Few arguments"); }
+  do {
+    POP(y, args);
+    do_compare(x, y, r, <);
+    x = y;
+  } while (PAIRP(args) && r);
+  return r ? T : F;
 }
 static sen_obj *
 nf_gre(sen_ctx *ctx, sen_obj *args, sen_ql_co *co)
 {
-  register cell *x = CAR(args);
-  register cell *y = CADR(args);
-  do_compare(x, y, >);
+  int r = 1;
+  register cell *x, *y;
+  POP(x, args);
+  if (!PAIRP(args)) { QLERR("Few arguments"); }
+  do {
+    POP(y, args);
+    do_compare(x, y, r, >);
+    x = y;
+  } while (PAIRP(args) && r);
+  return r ? T : F;
 }
 static sen_obj *
 nf_leq(sen_ctx *ctx, sen_obj *args, sen_ql_co *co)
 {
-  register cell *x = CAR(args);
-  register cell *y = CADR(args);
-  do_compare(x, y, <=);
+  int r = 1;
+  register cell *x, *y;
+  POP(x, args);
+  if (!PAIRP(args)) { QLERR("Few arguments"); }
+  do {
+    POP(y, args);
+    do_compare(x, y, r, <=);
+    x = y;
+  } while (PAIRP(args) && r);
+  return r ? T : F;
 }
 static sen_obj *
 nf_geq(sen_ctx *ctx, sen_obj *args, sen_ql_co *co)
 {
-  register cell *x = CAR(args);
-  register cell *y = CADR(args);
-  do_compare(x, y, >=);
+  int r = 1;
+  register cell *x, *y;
+  POP(x, args);
+  if (!PAIRP(args)) { QLERR("Few arguments"); }
+  do {
+    POP(y, args);
+    do_compare(x, y, r, >=);
+    x = y;
+  } while (PAIRP(args) && r);
+  return r ? T : F;
 }
 static sen_obj *
 nf_symbol(sen_ctx *ctx, sen_obj *args, sen_ql_co *co)
@@ -2587,7 +2688,7 @@ static sen_obj *
 nf_tonumber(sen_ctx *ctx, sen_obj *args, sen_ql_co *co)
 {
   sen_obj *x, *v;
-  if (!PAIRP(args)) { QLERR("list expected"); }
+  if (!PAIRP(args)) { QLERR("list required"); }
   x = CAR(args);
   switch (x->type) {
   case sen_ql_bulk :
@@ -2615,14 +2716,15 @@ nf_totime(sen_ctx *ctx, sen_obj *args, sen_ql_co *co)
 {
   sen_timeval tv;
   sen_obj *x, *v;
-  if (!PAIRP(args)) { QLERR("list expected"); }
+  if (!PAIRP(args)) { QLERR("list required"); }
   x = CAR(args);
   switch (x->type) {
   case sen_ql_bulk :
     {
-      const char *fmt = NULL;
+      /*
       if (PAIRP(CDR(args)) && BULKP(CADR(args))) { fmt = STRVALUE(CADR(args)); }
-      if (sen_str2timeval(STRVALUE(x), x->u.b.size, fmt, &tv)) {
+      */
+      if (sen_str2timeval(STRVALUE(x), x->u.b.size, &tv)) {
         QLERR("cast error");
       }
       SEN_OBJ_NEW(ctx, v);
@@ -2649,6 +2751,44 @@ nf_totime(sen_ctx *ctx, sen_obj *args, sen_ql_co *co)
   }
   return v;
 }
+static sen_obj *
+nf_substrb(sen_ctx *ctx, sen_obj *args, sen_ql_co *co)
+{
+  sen_obj *str, *s, *e;
+  int64_t is, ie;
+  if (!PAIRP(args)) { QLERR("list required"); }
+  POP(str, args);
+  if (!BULKP(str)) { QLERR("string required"); }
+  POP(s, args);
+  if (!INTP(s)) { QLERR("integer required"); }
+  POP(e, args);
+  if (!INTP(e)) { QLERR("integer required"); }
+  is = IVALUE(s);
+  ie = IVALUE(e) + 1;
+  if (ie <= 0) {
+    ie = str->u.b.size + ie;
+    if (ie < 0) { ie = 0; }
+  } else if (ie > str->u.b.size) {
+    ie = str->u.b.size;
+  }
+  if (is < 0) {
+    is = str->u.b.size + is + 1;
+    if (is < 0) { is = 0; }
+  } else if (is > str->u.b.size) {
+    is = str->u.b.size;
+  }
+  if (is < ie) {
+    return sen_ql_mk_string(ctx, STRVALUE(str) + is, ie - is);
+  } else {
+    sen_obj *o;
+    SEN_OBJ_NEW(ctx, o);
+    o->flags = 0;
+    o->type = sen_ql_bulk;
+    o->u.b.size = 0;
+    o->u.b.value = NULL;
+    return o;
+  }
+}
 
 /* ========== Initialization of internal keywords ========== */
 
@@ -2656,7 +2796,7 @@ inline static void
 mk_syntax(sen_ctx *ctx, uint8_t op, char *name)
 {
   cell *x;
-  if ((x = sen_ql_mk_symbol(ctx, name)) != F) {
+  if ((x = INTERN(name)) != F) {
     x->type = sen_ql_syntax;
     SYNTAXNUM(x) = op;
   }
@@ -2666,7 +2806,7 @@ inline static void
 mk_proc(sen_ctx *ctx, uint8_t op, char *name)
 {
   cell *x;
-  if ((x = sen_ql_mk_symbol(ctx, name)) != F) {
+  if ((x = INTERN(name)) != F) {
     x->type = sen_ql_proc;
     IVALUE(x) = (int64_t) op;
   }
@@ -2697,7 +2837,7 @@ init_vars_global(sen_ctx *ctx)
   /* init global_env */
   ctx->global_env = CONS(NIL, NIL);
   /* init else */
-  if ((x = sen_ql_mk_symbol(ctx, "else")) != F) {
+  if ((x = INTERN("else")) != F) {
     CAR(ctx->global_env) = CONS(CONS(x, T), CAR(ctx->global_env));
   }
 }
@@ -2741,6 +2881,7 @@ init_procs(sen_ctx *ctx)
   mk_proc(ctx, OP_PUT, "put");
   mk_proc(ctx, OP_GET, "get");
   mk_proc(ctx, OP_QUIT, "quit");
+  mk_proc(ctx, OP_SDOWN, "shutdown");
   sen_ql_def_native_func(ctx, "+", nf_add);
   sen_ql_def_native_func(ctx, "-", nf_sub);
   sen_ql_def_native_func(ctx, "*", nf_mul);
@@ -2788,6 +2929,7 @@ init_procs(sen_ctx *ctx)
   sen_ql_def_native_func(ctx, "timestr", nf_timestr);
   sen_ql_def_native_func(ctx, "x->time", nf_totime);
   sen_ql_def_native_func(ctx, "x->number", nf_tonumber);
+  sen_ql_def_native_func(ctx, "substrb", nf_substrb);
 }
 
 /* initialize several globals */
