@@ -4,12 +4,12 @@
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.
-  
+
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
-  
+
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
@@ -17,7 +17,6 @@
 #include "senna_in.h"
 #include <string.h>
 #include <limits.h>
-#include "str.h"
 #include "set.h"
 
 #define INITIAL_N_ENTRIES 256U
@@ -29,6 +28,8 @@
 
 #define GARBAGE ((entry *) 1)
 
+#define STEP(x) (((x) >> 2) | 0x1010101)
+
 typedef struct _sen_set_element entry;
 typedef struct _sen_set_element_str entry_str;
 
@@ -38,6 +39,7 @@ inline static entry *
 entry_new(sen_set *set)
 {
   entry *e;
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
   if (set->garbages) {
     e = set->garbages;
     set->garbages = *((entry **)e);
@@ -62,8 +64,9 @@ sen_set *
 sen_set_open(uint32_t key_size, uint32_t value_size, uint32_t init_size)
 {
   sen_set *set;
-  uint32_t entry_size, i, mod;
-  for (i = INITIAL_INDEX_SIZE; i < init_size; i *= 2);
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
+  uint32_t entry_size, n, mod;
+  for (n = INITIAL_INDEX_SIZE; n < init_size; n *= 2);
   switch (key_size) {
   case 0 :
     entry_size = (intptr_t)(&((entry_str *)0)->dummy) + value_size;
@@ -83,8 +86,8 @@ sen_set_open(uint32_t key_size, uint32_t value_size, uint32_t init_size)
   set->key_size = key_size;
   set->value_size = value_size;
   set->entry_size = entry_size;
-  set->max_offset = i - 1;
-  if (!(set->index = SEN_CALLOC(i * sizeof(entry *)))) {
+  set->max_offset = n - 1;
+  if (!(set->index = SEN_CALLOC(n * sizeof(entry *)))) {
     SEN_FREE(set);
     return NULL;
   }
@@ -94,40 +97,40 @@ sen_set_open(uint32_t key_size, uint32_t value_size, uint32_t init_size)
 sen_rc
 sen_set_reset(sen_set * set, uint32_t ne)
 {
-  uint32_t i, j, m, n;
+  uint32_t i, j, m, n, s;
   entry **index, *e, **sp, **dp;
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
   if (!ne) { ne = set->n_entries * 2; }
   if (ne > INT_MAX) { return sen_memory_exhausted; }
   for (n = INITIAL_INDEX_SIZE; n <= ne; n *= 2);
   if (!(index = SEN_CALLOC(n * sizeof(entry *)))) { return sen_memory_exhausted; }
   m = n - 1;
   if (set->key_size) {
-    for (j = set->max_offset + 1, sp = set->index; j; j--, sp++) { 
-      e = *sp; 
-      if (!e || (e == GARBAGE)) { continue; } 
-      for (i = e->key; ; i += 1048573) { 
-        dp = index + (i & m); 
-        if (!*dp) { break; } 
-      } 
-      *dp = e; 
-    } 
+    for (j = set->max_offset + 1, sp = set->index; j; j--, sp++) {
+      e = *sp;
+      if (!e || (e == GARBAGE)) { continue; }
+      for (i = e->key, s = STEP(i); ; i += s) {
+        dp = index + (i & m);
+        if (!*dp) { break; }
+      }
+      *dp = e;
+    }
   } else {
-    for (j = set->max_offset + 1, sp = set->index; j; j--, sp++) { 
-      e = *sp; 
-      if (!e || (e == GARBAGE)) { continue; } 
-      for (i = SEN_SET_STRHASH(e); ; i += 1048573) { 
-        dp = index + (i & m); 
-        if (!*dp) { break; } 
-      } 
-      *dp = e; 
-    } 
+    for (j = set->max_offset + 1, sp = set->index; j; j--, sp++) {
+      e = *sp;
+      if (!e || (e == GARBAGE)) { continue; }
+      for (i = SEN_SET_STRHASH(e), s = STEP(i); ; i += s) {
+        dp = index + (i & m);
+        if (!*dp) { break; }
+      }
+      *dp = e;
+    }
   }
   {
     entry **i0 = set->index;
     set->index = index;
     set->max_offset = m;
     set->n_garbages = 0;
-    set->garbages = NULL;
     SEN_FREE(i0);
   }
   return sen_success;
@@ -137,6 +140,7 @@ sen_rc
 sen_set_close(sen_set * set)
 {
   uint32_t i;
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
   if (!set) { return sen_invalid_argument; }
   if (!set->key_size) {
     entry *e, **sp;
@@ -185,8 +189,8 @@ sen_set_eh *
 sen_set_int_at(sen_set *set, const uint32_t *key, void **value)
 {
   entry *e, **ep, **index = set->index;
-  uint32_t h = *key, i, m = set->max_offset;
-  for (i = h; ep = index + (i & m), (e = *ep); i += 1048573) {
+  uint32_t h = *key, i, m = set->max_offset, s = STEP(h);
+  for (i = h; ep = index + (i & m), (e = *ep); i += s) {
     if (e == GARBAGE) { continue; }
     if (e->key == h) {
       if (value) { *value = e->dummy; }
@@ -200,8 +204,8 @@ sen_set_eh *
 sen_set_str_at(sen_set *set, const char *key, void **value)
 {
   entry *e, **ep, **index = set->index;
-  uint32_t h = str_hash((unsigned char *)key), i, m = set->max_offset;
-  for (i = h; ep = index + (i & m), (e = *ep); i += 1048573) {
+  uint32_t h = str_hash((unsigned char *)key), i, m = set->max_offset, s = STEP(h);
+  for (i = h; ep = index + (i & m), (e = *ep); i += s) {
     if (e == GARBAGE) { continue; }
     if (SEN_SET_STRHASH(e) == h && !strcmp(key, SEN_SET_STRKEY(e))) {
       if (value) { *value = SEN_SET_STRVAL(e); }
@@ -215,8 +219,8 @@ sen_set_eh *
 sen_set_bin_at(sen_set *set, const void *key, void **value)
 {
   entry *e, **ep, **index = set->index;
-  uint32_t h = bin_hash(key, set->key_size), i, m = set->max_offset;
-  for (i = h; ep = index + (i & m), (e = *ep); i += 1048573) {
+  uint32_t h = bin_hash(key, set->key_size), i, m = set->max_offset, s = STEP(h);
+  for (i = h; ep = index + (i & m), (e = *ep); i += s) {
     if (e == GARBAGE) { continue; }
     if (e->key == h && !memcmp(key, e->dummy, set->key_size)) {
       if (value) { *value = SEN_SET_BINVAL(e, set); }
@@ -230,7 +234,7 @@ sen_set_eh *
 sen_set_at(sen_set *set, const void *key, void **value)
 {
   if (set->arrayp) {
-    sen_set_reset(set, 0);
+    if (sen_set_reset(set, 0)) { return NULL; }
     set->curr_entry = 0;
     set->arrayp = 0;
   }
@@ -247,20 +251,30 @@ sen_set_at(sen_set *set, const void *key, void **value)
 inline static sen_set_eh *
 sen_set_int_get(sen_set *set, const uint32_t *key, void **value)
 {
+  static int _ncalls = 0, _ncolls = 0;
   entry *e, **ep, **np = NULL, **index = set->index;
-  uint32_t h = *key, i, m = set->max_offset;
-  for (i = h; ep = index + (i & m), e = *ep; i += 1048573) {
+  uint32_t h = *key, i, m = set->max_offset, s = STEP(h);
+  _ncalls++;
+  for (i = h; ep = index + (i & m), e = *ep; i += s) {
     if (e == GARBAGE) {
       if (!np) { np = ep; }
     } else {
       if (e->key == h) { goto exit; }
     }
+    if (!(++_ncolls % 1000000) && (_ncolls > _ncalls)) {
+      if (_ncolls < 0 || _ncalls < 0) {
+        _ncolls = 0; _ncalls = 0;
+      } else {
+        SEN_LOG(sen_log_notice, "total:%d collisions:%d nentries:%d maxoffset:%d",
+                _ncalls, _ncolls, set->n_entries, set->max_offset);
+      }
+    }
   }
   if (np) {
     set->n_garbages--;
     ep = np;
   }
-  e = entry_new(set);
+  if (!(e = entry_new(set))) { return NULL; }
   e->key = h;
   *ep = e;
   set->n_entries++;
@@ -272,24 +286,32 @@ exit :
 inline static sen_set_eh *
 sen_set_str_get(sen_set *set, const char *key, void **value)
 {
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
   entry *e, **ep, **np = NULL, **index = set->index;
-  uint32_t h = str_hash((unsigned char *)key), i, m = set->max_offset;
-  for (i = h; ep = index + (i & m), e = *ep; i += 1048573) {
+  uint32_t h = str_hash((unsigned char *)key), i, m = set->max_offset, s = STEP(h);
+  for (i = h; ep = index + (i & m), e = *ep; i += s) {
     if (e == GARBAGE) {
       if (!np) { np = ep; }
     } else {
       if (SEN_SET_STRHASH(e) == h && !strcmp(key, SEN_SET_STRKEY(e))) { goto exit; }
     }
   }
-  if (np) {
-    set->n_garbages--;
-    ep = np;
+  {
+    char *keybuf = SEN_STRDUP(key);
+    if (!keybuf) { return NULL; }
+    if (!(e = entry_new(set))) {
+      SEN_FREE(keybuf);
+      return NULL;
+    }
+    SEN_SET_STRHASH(e) = h;
+    SEN_SET_STRKEY(e) = keybuf;
+    if (np) {
+      set->n_garbages--;
+      ep = np;
+    }
+    *ep = e;
+    set->n_entries++;
   }
-  e = entry_new(set);
-  SEN_SET_STRHASH(e) = h;
-  SEN_SET_STRKEY(e) = SEN_STRDUP(key);
-  *ep = e;
-  set->n_entries++;
 exit :
   if (value) { *value = SEN_SET_STRVAL(e); }
   return ep;
@@ -299,8 +321,8 @@ inline static sen_set_eh *
 sen_set_bin_get(sen_set *set, const void *key, void **value)
 {
   entry *e, **ep, **np = NULL, **index = set->index;
-  uint32_t h = bin_hash(key, set->key_size), i, m = set->max_offset;
-  for (i = h; ep = index + (i & m), e = *ep; i += 1048573) {
+  uint32_t h = bin_hash(key, set->key_size), i, m = set->max_offset, s = STEP(h);
+  for (i = h; ep = index + (i & m), e = *ep; i += s) {
     if (e == GARBAGE) {
       if (!np) { np = ep; }
     } else {
@@ -313,7 +335,7 @@ sen_set_bin_get(sen_set *set, const void *key, void **value)
     set->n_garbages--;
     ep = np;
   }
-  e = entry_new(set);
+  if (!(e = entry_new(set))) { return NULL; }
   e->key = h;
   memcpy(e->dummy, key, set->key_size);
   *ep = e;
@@ -326,8 +348,9 @@ exit :
 sen_set_eh *
 sen_set_get(sen_set *set, const void *key, void **value)
 {
+  if (!set) { return NULL; }
   if (set->arrayp) {
-    sen_set_reset(set, 0);
+    if (sen_set_reset(set, 0)) { return NULL; }
     set->curr_entry = 0;
     set->arrayp = 0;
   } else if ((set->n_entries + set->n_garbages) * 2 > set->max_offset) {
@@ -347,11 +370,11 @@ sen_rc
 sen_set_del(sen_set *set, sen_set_eh *ep)
 {
   entry *e;
-  if (!ep || !*ep) { return sen_invalid_argument; }
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
+  if (!set || !ep || !*ep) { return sen_invalid_argument; }
   e = *ep;
   *ep = GARBAGE;
   if (!set->key_size) { SEN_FREE(SEN_SET_STRKEY(e)); }
-  //  memset(e, 0, set->entry_size);
   *((entry **)e) = set->garbages;
   set->garbages = e;
   set->n_entries--;
@@ -363,6 +386,8 @@ sen_set_cursor *
 sen_set_cursor_open(sen_set *set)
 {
   sen_set_cursor *c;
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
+  if (!set) { return NULL; }
   if (!(c = SEN_MALLOC(sizeof(sen_set_cursor)))) { return NULL; }
   c->set = set;
   c->index = set->index;
@@ -411,6 +436,7 @@ sen_set_cursor_next(sen_set_cursor *c, void **key, void **value)
 sen_rc
 sen_set_cursor_close(sen_set_cursor *cursor)
 {
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
   SEN_FREE(cursor);
   return sen_success;
 }
@@ -605,6 +631,7 @@ func_bin(sen_set *sa, entry **a, sen_set *sb, entry **b, void *arg)
 sen_set_eh *
 sen_set_sort(sen_set *set, int limit, sen_set_sort_optarg *optarg)
 {
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
   entry **res;
   int dir = 1;
   if (!set) {
@@ -629,7 +656,7 @@ sen_set_sort(sen_set *set, int limit, sen_set_sort_optarg *optarg)
     } else if (optarg->compar_arg) {
       sort_int(set, res, limit, ((intptr_t)optarg->compar_arg) / sizeof(int32_t), dir);
       goto exit;
-    } 
+    }
   }
   switch (set->key_size) {
   case 0 :
@@ -693,8 +720,7 @@ sen_set_union(sen_set *a, sen_set *b)
       if (sen_set_at(a, key, &va)) {
         /* do copy of merge? */
       } else {
-        sen_set_get(a, key, &va);
-        memcpy(va, vb, value_size);
+        if (sen_set_get(a, key, &va)) { memcpy(va, vb, value_size); }
       }
       i--;
     }
@@ -793,7 +819,9 @@ sen_set_difference(sen_set *a, sen_set *b)
 sen_rc
 sen_set_array_init(sen_set *set, uint32_t size)
 {
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
   SEN_ASSERT(!set->n_entries);
+  SEN_ASSERT(!set->garbages);
   set->arrayp = 1;
   set->curr_entry = 0;
   if (set->chunks[SEN_SET_MAX_CHUNK]) {
