@@ -100,6 +100,10 @@ sen_bm_preBmBc(const unsigned char *x, size_t m, size_t *bmBc)
     } else { \
       cond->last_found = found; \
     } \
+    if (flags & SEN_SNIP_SKIP_LEADING_SPACES) { \
+      while (offset < object->orig_blen && \
+             (i = sen_isspace(object->orig + offset, object->encoding))) { offset += i; } \
+    } \
     cond->start_offset = offset; \
     for (i = cond->last_found; i < found + m; i++) { \
       if (object->checks[i] > 0) { \
@@ -126,7 +130,7 @@ sen_bm_preBmBc(const unsigned char *x, size_t m, size_t *bmBc)
 }
 
 void
-sen_bm_tunedbm(snip_cond *cond, sen_nstr *object)
+sen_bm_tunedbm(snip_cond *cond, sen_nstr *object, int flags)
 {
   register unsigned char *limit, ck;
   register const unsigned char *p, *cp;
@@ -310,9 +314,12 @@ sen_snip_add_cond(sen_snip *snip,
   if (opentag) {
     if (snip->flags & SEN_SNIP_COPY_TAG) {
       char *t = SEN_MALLOC(opentag_len + 1);
-      if (!t) { return sen_memory_exhausted; }
+      if (!t) {
+        sen_snip_cond_close(cond);
+        return sen_memory_exhausted;
+      }
       memcpy(t, opentag, opentag_len);
-      t[opentag_len]= '\0';
+      t[opentag_len]= '\0'; /* not required, but for ql use */
       cond->opentag = t;
     } else {
       cond->opentag = opentag;
@@ -327,7 +334,7 @@ sen_snip_add_cond(sen_snip *snip,
       char *t = SEN_MALLOC(closetag_len + 1);
       if (!t) { return sen_memory_exhausted; }
       memcpy(t, closetag, closetag_len);
-      t[closetag_len]= '\0';
+      t[closetag_len]= '\0'; /* not required, but for ql use */
       cond->closetag = t;
     } else {
       cond->closetag = closetag;
@@ -385,8 +392,30 @@ sen_snip_open(sen_encoding encoding, int flags, unsigned int width,
   ret->flags = flags;
   ret->width = width;
   ret->max_results = max_results;
-  ret->defaultopentag = defaultopentag;
-  ret->defaultclosetag = defaultclosetag;
+  if (flags & SEN_SNIP_COPY_TAG) {
+    char *t;
+    t = SEN_MALLOC(defaultopentag_len + 1);
+    if (!t) {
+      SEN_FREE(ret);
+      return NULL;
+    }
+    memcpy(t, defaultopentag, defaultopentag_len);
+    t[defaultopentag_len]= '\0'; /* not required, but for ql use */
+    ret->defaultopentag = t;
+
+    t = SEN_MALLOC(defaultclosetag_len + 1);
+    if (!t) {
+      SEN_FREE((void *)ret->defaultopentag);
+      SEN_FREE(ret);
+      return NULL;
+    }
+    memcpy(t, defaultclosetag, defaultclosetag_len);
+    t[defaultclosetag_len]= '\0'; /* not required, but for ql use */
+    ret->defaultclosetag = t;
+  } else {
+    ret->defaultopentag = defaultopentag;
+    ret->defaultclosetag = defaultclosetag;
+  }
   ret->defaultopentag_len = defaultopentag_len;
   ret->defaultclosetag_len = defaultclosetag_len;
   ret->cond_len = 0;
@@ -468,7 +497,7 @@ sen_snip_exec(sen_snip *snip, const char *string, unsigned int string_len,
     return sen_memory_exhausted;
   }
   for (i = 0; i < snip->cond_len; i++) {
-    sen_bm_tunedbm(snip->cond + i, snip->nstr);
+    sen_bm_tunedbm(snip->cond + i, snip->nstr, snip->flags);
   }
 
   {
@@ -516,7 +545,7 @@ sen_snip_exec(sen_snip *snip, const char *string, unsigned int string_len,
               }
             }
             if (exclude_other_cond) {
-              sen_bm_tunedbm(cond, snip->nstr);
+              sen_bm_tunedbm(cond, snip->nstr, snip->flags);
               continue;
             }
           }
@@ -529,7 +558,7 @@ sen_snip_exec(sen_snip *snip, const char *string, unsigned int string_len,
           /* check nesting to make valid HTML */
           /* ToDo: allow <test><te>te</te><st>st</st></test> */
           if (cond->start_offset < last_tag_end) {
-            sen_bm_tunedbm(cond, snip->nstr);
+            sen_bm_tunedbm(cond, snip->nstr, snip->flags);
             continue;
           }
         }
@@ -537,7 +566,7 @@ sen_snip_exec(sen_snip *snip, const char *string, unsigned int string_len,
           /* If a keyword gets across a snippet, */
           /* it was skipped and never to be tagged. */
           cond->stopflag = SNIPCOND_ACROSS;
-          sen_bm_tunedbm(cond, snip->nstr);
+          sen_bm_tunedbm(cond, snip->nstr, snip->flags);
         } else {
           found_cond = 1;
           if (cond->count == 0) {
@@ -556,7 +585,7 @@ sen_snip_exec(sen_snip *snip, const char *string, unsigned int string_len,
           if (++snip->tag_count >= MAX_SNIP_TAG_COUNT) {
             break;
           }
-          sen_bm_tunedbm(cond, snip->nstr);
+          sen_bm_tunedbm(cond, snip->nstr, snip->flags);
         }
       }
       if (!found_cond) {
