@@ -205,6 +205,11 @@ sen_ra_at(sen_ra *ra, sen_id id)
 
 /**** jagged arrays ****/
 
+#define SEN_JA_MAX_CACHE (4294967295U)
+#define SEG_NOT_ASSIGNED 0xffffffff
+
+#ifdef USE_JA01
+
 #define USE_SEG_MAP
 
 #define SEN_JA_IDSTR "SENNA:JA:01.000"
@@ -226,7 +231,7 @@ sen_ra_at(sen_ra *ra, sen_id id)
 
 #define JA_N_ESEGMENTS (1U << 9)
 
-#define SEN_JA_MAX_CACHE (4294967295U)
+typedef struct _sen_ja_einfo sen_ja_einfo;
 
 struct _sen_ja_einfo {
   union {
@@ -259,6 +264,11 @@ typedef struct {
   uint32_t pos;
 } ja_pos;
 
+struct _sen_ja {
+  sen_io *io;
+  struct sen_ja_header *header;
+};
+
 struct sen_ja_header {
   char idstr[16];
   unsigned max_element_size;
@@ -272,7 +282,6 @@ struct sen_ja_header {
 
 #define JA_SEG_ESEG 1;
 #define JA_SEG_BSEG 2;
-#define SEG_NOT_ASSIGNED 0xffffffff
 
 sen_ja *
 sen_ja_create(const char *path, unsigned int max_element_size)
@@ -577,11 +586,357 @@ sen_ja_replace(sen_ja *ja, sen_id id, sen_ja_einfo *ei)
   if (!einfo) { return sen_memory_exhausted; }
   eback = einfo[pos];
   einfo[pos] = *ei;
-  // SEN_SET_64BIT(&einfo[pos], *ei);
+  SEN_SET_64BIT(&einfo[pos], *ei);
   sen_ja_free(ja, &eback);
   return sen_success;
 }
 
+#else /* USE_JA01 */
+
+#define SEN_JA_IDSTR "SENNA:JA:02.000"
+
+struct _sen_ja {
+  sen_io *io;
+  struct sen_ja_header *header;
+  uint32_t *dsegs;
+  uint32_t *esegs;
+};
+
+struct sen_ja_header {
+  char idstr[16];
+  uint32_t flags;
+  uint32_t align_width;
+  uint32_t segment_width;
+  uint32_t element_size;
+  uint32_t curr_pos;
+};
+
+#define SEN_JA_DEFAULT_ALIGN_WIDTH    4
+#define SEN_JA_DEFAULT_SEGMENT_WIDTH 20
+
+#define JA_SEGMENT_SIZE (1 << SEN_JA_DEFAULT_SEGMENT_WIDTH)
+
+#define JA_ALIGN_WIDTH (ja->header->align_width)
+#define JA_ALIGN_MASK ((1U << JA_ALIGN_WIDTH) - 1)
+
+#define JA_DSEG_WIDTH (ja->header->segment_width)
+#define JA_DSEG_MASK ((1U << JA_DSEG_WIDTH) - 1)
+
+#define JA_ESEG_WIDTH (JA_DSEG_WIDTH - 3)
+#define JA_ESEG_MASK ((1U << JA_ESEG_WIDTH) - 1)
+
+#define JA_EPOS_WIDTH (JA_DSEG_WIDTH - JA_ALIGN_WIDTH)
+#define JA_EPOS_MASK ((1U << JA_EPOS_WIDTH) - 1)
+
+// segment_width, align_width, flags
+sen_ja *
+sen_ja_create(const char *path, unsigned int max_element_size)
+{
+  int i;
+  sen_io *io;
+  sen_ja *ja;
+  struct sen_ja_header *header;
+  uint32_t align_width = SEN_JA_DEFAULT_ALIGN_WIDTH;
+  uint32_t segment_width = SEN_JA_DEFAULT_SEGMENT_WIDTH;
+  uint32_t max_dsegs = 1 << (32 - (segment_width - align_width));
+  uint32_t max_esegs = 1 << (31 - segment_width);
+  if (align_width > segment_width || align_width + 32 < segment_width) {
+    SEN_LOG(sen_log_error, "invalid align_width, segment_width value");
+    return NULL;
+  }
+  io = sen_io_create(path,
+                     sizeof(struct sen_ja_header) +
+                     max_dsegs * sizeof(int32_t) +
+                     max_esegs * sizeof(int32_t),
+                     1 << segment_width, max_dsegs, sen_io_auto, SEN_JA_MAX_CACHE);
+  if (!io) { return NULL; }
+  header = sen_io_header(io);
+  memcpy(header->idstr, SEN_JA_IDSTR, 16);
+  header->flags = 0;
+  header->align_width = align_width;
+  header->segment_width = segment_width;
+  header->element_size = max_element_size;
+  header->curr_pos = 0;
+  if (!(ja = SEN_GMALLOC(sizeof(sen_ja)))) {
+    sen_io_close(io);
+    return NULL;
+  }
+  ja->io = io;
+  ja->header = header;
+  ja->dsegs = (uint32_t *)(((uintptr_t) header) + sizeof(struct sen_ja_header));
+  ja->esegs = &ja->dsegs[max_dsegs];
+  for (i = 0; i < max_esegs; i++) { ja->esegs[i] = SEG_NOT_ASSIGNED; }
+  return ja;
+}
+
+sen_ja *
+sen_ja_open(const char *path)
+{
+  sen_io *io;
+  sen_ja *ja = NULL;
+  struct sen_ja_header *header;
+  io = sen_io_open(path, sen_io_auto, SEN_JA_MAX_CACHE);
+  if (!io) { return NULL; }
+  header = sen_io_header(io);
+  if (memcmp(header->idstr, SEN_JA_IDSTR, 16)) {
+    SEN_LOG(sen_log_error, "ja_idstr (%s)", header->idstr);
+    sen_io_close(io);
+    return NULL;
+  }
+  if (!(ja = SEN_GMALLOC(sizeof(sen_ja)))) {
+    sen_io_close(io);
+    return NULL;
+  }
+  ja->io = io;
+  ja->header = header;
+  ja->dsegs = (uint32_t *)(((uintptr_t) header) + sizeof(struct sen_ja_header));
+  ja->esegs = &ja->dsegs[1 << (32 - JA_EPOS_WIDTH)];
+  return ja;
+}
+
+sen_rc
+sen_ja_info(sen_ja *ja, unsigned int *max_element_size)
+{
+  if (!ja) { return sen_invalid_argument; }
+  if (max_element_size) { *max_element_size = 4294967295U; }
+  return sen_success;
+}
+
+sen_rc
+sen_ja_close(sen_ja *ja)
+{
+  sen_rc rc;
+  if (!ja) { return sen_invalid_argument; }
+  rc = sen_io_close(ja->io);
+  SEN_GFREE(ja);
+  return rc;
+}
+
+sen_rc
+sen_ja_remove(const char *path)
+{
+  if (!path) { return sen_invalid_argument; }
+  return sen_io_remove(path);
+}
+
+void *
+sen_ja_ref(sen_ja *ja, sen_id id, uint32_t *value_len)
+{
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
+  sen_io_ja_einfo *ep;
+  uint32_t lseg = id >> JA_ESEG_WIDTH;
+  uint32_t lpos = id & JA_ESEG_MASK;
+  uint32_t *pseg = &ja->esegs[lseg];
+  if (*pseg == SEG_NOT_ASSIGNED) { *value_len = 0; return NULL; }
+  SEN_IO_SEG_MAP(ja->io, *pseg, ep);
+  if (!ep) { *value_len = 0; return NULL; }
+  ep += lpos;
+  for (;;) {
+    void *value;
+    uint32_t epos = ep->pos;
+    uint32_t dseg = epos >> JA_EPOS_WIDTH;
+    uint32_t doff = (epos & JA_EPOS_MASK) << JA_ALIGN_WIDTH;
+    if (!(*value_len = ep->size)) { return NULL; }
+    if (sen_io_read_ja(ja->io, ctx, ep, epos, id, dseg, doff, &value, value_len)
+        != sen_internal_error) {
+      return value;
+    }
+  }
+}
+
+sen_rc
+sen_ja_unref(sen_ja *ja, sen_id id, void *value, uint32_t value_len)
+{
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
+  SEN_FREE((void *)(((uintptr_t) value) - sizeof(sen_io_ja_ehead)));
+  return sen_success;
+}
+
+#define SEG_ESEG 0xffffffff
+
+sen_rc
+assign_eseg(sen_ja *ja, uint32_t lseg)
+{
+  int i;
+  uint32_t max_dsegs = 1 << (32 - JA_EPOS_WIDTH);
+  for (i = 0; i < max_dsegs; i++) {
+    if (ja->dsegs[i] == 0) {
+      if (ja->header->curr_pos && i == (ja->header->curr_pos >> JA_EPOS_WIDTH)) {
+        // todo : curr_pos must be moved.
+      }
+      ja->dsegs[i] = SEG_ESEG;
+      ja->esegs[lseg] = i;
+      return sen_success;
+    }
+  }
+  return sen_internal_error;
+}
+
+static sen_rc
+assign_rec(sen_ja *ja, int value_len)
+{
+  if (ja->header->curr_pos) {
+    uint32_t doff = (ja->header->curr_pos & JA_EPOS_MASK) << JA_ALIGN_WIDTH;
+    if (doff + value_len + sizeof(sen_io_ja_ehead) <= (1 << JA_DSEG_WIDTH)) {
+      return sen_success;
+    }
+  }
+  {
+    uint32_t max_dsegs = 1 << (32 - JA_EPOS_WIDTH);
+    /* if (value_len <= JA_DSEG_MASK) { todo : } else */
+    {
+      uint32_t i, j;
+      uint32_t nseg = (value_len + sizeof(sen_io_ja_ehead) + JA_DSEG_MASK) >> JA_DSEG_WIDTH;
+      for (i = 0, j = 0; i < max_dsegs; i++) {
+        if (ja->dsegs[i] == 0) {
+          if (++j == nseg) {
+            ja->header->curr_pos = (i + 1 - j) << JA_EPOS_WIDTH;
+            return sen_success;
+          }
+        } else {
+          j = 0;
+        }
+      }
+    }
+    return sen_other_error;
+  }
+}
+
+sen_rc
+sen_ja_put(sen_ja *ja, sen_id id, void *value, int value_len, int flags)
+{
+  sen_rc rc = sen_success;
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
+  uint32_t newpos = 0;
+  sen_io_ja_einfo *ep;
+  {
+    uint32_t lseg = id >> JA_ESEG_WIDTH;
+    uint32_t lpos = id & JA_ESEG_MASK;
+    uint32_t *pseg = &ja->esegs[lseg];
+    if (*pseg == SEG_NOT_ASSIGNED && (rc = assign_eseg(ja, lseg))) { return rc; }
+    SEN_IO_SEG_MAP(ja->io, *pseg, ep);
+    if (!ep) { return sen_other_error; }
+    ep += lpos;
+  }
+  if (value_len) {
+    uint32_t dseg, doff, inc;
+    if ((rc = assign_rec(ja, value_len))) { return rc; }
+    newpos = ja->header->curr_pos;
+    dseg = newpos >> JA_EPOS_WIDTH;
+    doff = (newpos & JA_EPOS_MASK) << JA_ALIGN_WIDTH;
+    rc = sen_io_write_ja(ja->io, ctx, id, dseg, doff, value, value_len);
+    if (rc) { return rc; }
+    inc = ((value_len + sizeof(sen_io_ja_ehead) + JA_ALIGN_MASK) >> JA_ALIGN_WIDTH);
+    ja->header->curr_pos = ((newpos + inc) & JA_EPOS_MASK) ? (newpos + inc) : 0;
+    {
+      uint32_t max = 1U << JA_EPOS_WIDTH;
+      while (ja->dsegs[dseg] + inc > max) {
+        inc -= max - ja->dsegs[dseg];
+        ja->dsegs[dseg++] = max;
+      }
+      ja->dsegs[dseg] += inc;
+    }
+  }
+  {
+    sen_io_ja_einfo ne, oe;
+    oe = *ep;
+    ne.pos = newpos;
+    ne.size = value_len;
+    *ep = ne; // atomic_set64 maybe more suitable.
+    if (oe.size) {
+      uint32_t dseg, off, dec, max = 1U << JA_EPOS_WIDTH;
+      dseg = oe.pos >> JA_EPOS_WIDTH;
+      off = (oe.pos & JA_EPOS_MASK);
+      dec = ((oe.size + sizeof(sen_io_ja_ehead) + JA_ALIGN_MASK) >> JA_ALIGN_WIDTH);
+      if (off + dec > max) {
+        dec -= max - off;
+        ja->dsegs[dseg++] -= max - off;
+        while (dec > max) {
+          dec -= max;
+          ja->dsegs[dseg++] = 0;
+        }
+        {
+          uint32_t vsize = (dec << JA_ALIGN_WIDTH) - sizeof(sen_io_ja_ehead);
+          rc = sen_io_write_ja_ehead(ja->io, ctx, 0, dseg, 0, vsize);
+          // todo : handle rc..
+        }
+      }
+      ja->dsegs[dseg] -= dec;
+    }
+  }
+  return rc;
+}
+
+static sen_rc
+sen_ja_defrag_seg(sen_ja *ja, uint32_t dseg)
+{
+  sen_rc rc = sen_success;
+  sen_ctx *ctx = &sen_gctx; /* todo : replace it with the local ctx */
+  sen_io_win iw;
+  uint32_t epos = dseg << JA_EPOS_WIDTH;
+  uint32_t segsize = (1 << JA_DSEG_WIDTH);
+  byte *v = sen_io_win_map(ja->io, ctx, &iw, dseg, 0, segsize, sen_io_rdonly);
+  byte *ve = v + segsize;
+  uint32_t *segusage = &ja->dsegs[dseg];
+  if (!v) { return sen_internal_error; }
+  while (v < ve && *segusage) {
+    sen_io_ja_ehead *eh = (sen_io_ja_ehead *) v;
+    uint32_t inc = ((eh->size + sizeof(sen_io_ja_ehead) + JA_ALIGN_MASK) >> JA_ALIGN_WIDTH);
+    sen_io_ja_einfo *ep;
+    uint32_t lseg = eh->key >> JA_ESEG_WIDTH;
+    uint32_t lpos = eh->key & JA_ESEG_MASK;
+    uint32_t *pseg = &ja->esegs[lseg];
+    if (*pseg == SEG_NOT_ASSIGNED) {
+      SEN_LOG(sen_log_error, "ep is not assigned (%x)", lseg);
+      rc = sen_internal_error;
+      goto exit;
+    }
+    SEN_IO_SEG_MAP(ja->io, *pseg, ep);
+    if (!ep) {
+      SEN_LOG(sen_log_error, "ep map failed (%x)", lseg);
+      rc = sen_internal_error;
+      goto exit;
+    }
+    ep += lpos;
+    if (ep->pos == epos) {
+      if (ep->size != eh->size) {
+        SEN_LOG(sen_log_error, "ep->size(%d) != eh->size(%d)", ep->size, eh->size);
+        rc = sen_internal_error;
+        goto exit;
+      }
+      if ((rc = sen_ja_put(ja, eh->key, v + sizeof(sen_io_ja_ehead), eh->size, 0))) {
+        goto exit;
+      }
+    }
+    epos += inc;
+    v += (inc << JA_ALIGN_WIDTH);
+  }
+  if (*segusage) {
+    SEN_LOG(sen_log_error, "ja->dsegs[dseg] = %d after defrag", ja->dsegs[dseg]);
+    rc = sen_internal_error;
+  }
+exit :
+  sen_io_win_unmap(&iw);
+  return rc;
+}
+
+int
+sen_ja_defrag(sen_ja *ja, int threshold)
+{
+  int nsegs = 0;
+  uint32_t ts = 1U << (JA_EPOS_WIDTH - threshold);
+  uint32_t dseg, max_dsegs = 1 << (32 - JA_EPOS_WIDTH);
+  for (dseg = 0; dseg < max_dsegs; dseg++) {
+    if (dseg == ja->header->curr_pos >> JA_EPOS_WIDTH) { continue; }
+    if (ja->dsegs[dseg] && ja->dsegs[dseg] < ts) {
+      if (!sen_ja_defrag_seg(ja, dseg)) { nsegs++; }
+    }
+  }
+  return nsegs;
+}
+
+#endif /* USE_JA01 */
+
 /**** db ****/
 
 sen_rc
@@ -619,7 +974,9 @@ sen_db_store *
 sen_db_store_by_id(sen_db *s, sen_id id)
 {
   const char *name;
-  sen_db_store *slot = sen_array_at(&s->stores, id);
+  sen_db_store *slot;
+  if (!id) { return NULL; }
+  slot = sen_array_at(&s->stores, id);
   if (slot->type) { return slot; }
   if (!(name = _sen_sym_key(s->keys, id))) { return NULL; }
   return sen_db_store_open(s, name);
@@ -1079,11 +1436,7 @@ sen_db_store_remove(sen_db *s, const char *name)
     break;
   }
   sen_ja_unref(s->values, id, spec, spec_len);
-  {
-    sen_ja_einfo einfo;
-    memset(&einfo, 0, sizeof(sen_ja_einfo));
-    rc = sen_ja_replace(s->values, id, &einfo);
-  }
+  rc = sen_ja_put(s->values, id, NULL, 0, 0);
   rc = sen_sym_del(s->keys, name);
   MUTEX_UNLOCK(s->lock);
   sen_db_unlock(s);
@@ -1103,6 +1456,9 @@ sen_db_prepare_builtin_class(sen_db *db)
 {
   sen_db_store_spec spec;
   spec.type = sen_db_raw_class;
+  spec.n_triggers = 0;
+  spec.u.c.flags = 0;
+  spec.u.c.encoding = db->keys->encoding;
   /* 1 */
   spec.u.c.size = sizeof(int32_t);
   if (!sen_db_store_create(db, "<int>", &spec)) { return sen_memory_exhausted; }
