@@ -65,12 +65,6 @@ void *sen_ra_at(sen_ra *ra, sen_id id);
 /**** variable sized elements ****/
 
 typedef struct _sen_ja sen_ja;
-typedef struct _sen_ja_einfo sen_ja_einfo;
-
-struct _sen_ja {
-  sen_io *io;
-  struct sen_ja_header *header;
-};
 
 sen_ja *sen_ja_create(const char *path, unsigned int max_element_size);
 sen_ja *sen_ja_open(const char *path);
@@ -82,9 +76,7 @@ int sen_ja_at(sen_ja *ja, sen_id id, void *valbuf, int buf_size);
 
 void *sen_ja_ref(sen_ja *ja, sen_id id, uint32_t *value_len);
 sen_rc sen_ja_unref(sen_ja *ja, sen_id id, void *value, uint32_t value_len);
-
-int sen_ja_size(sen_ja *ja, sen_id id);
-sen_rc sen_ja_replace(sen_ja *ja, sen_id id, sen_ja_einfo *ei);
+int sen_ja_defrag(sen_ja *ja, int threshold);
 
 /**** db ****/
 
