@@ -96,7 +96,7 @@ sen_str_charlen_utf8(const unsigned char *str, const unsigned char *end)
   return 0;
 }
 
-size_t
+unsigned int
 sen_str_charlen(const char *str, sen_encoding encoding)
 {
   /* MEMO: This function requires null-terminated string as str.*/
@@ -1163,6 +1163,7 @@ sen_nstr_open(const char *str, size_t str_len, sen_encoding encoding, int flags)
   nstr->norm_blen = 0;
   nstr->checks = NULL;
   nstr->ctypes = NULL;
+  nstr->encoding = encoding;
   nstr->flags = flags;
   nstr->ctx = ctx;
   switch (encoding) {
@@ -1913,3 +1914,27 @@ sen_substring(char **str, char **str_end, int start, int end, sen_encoding encod
   }
   return sen_success;
 }
+
+int
+sen_str_normalize(const char *str, unsigned int str_len,
+                  sen_encoding encoding, int flags,
+                  char *nstrbuf, int buf_size)
+{
+  int len;
+  sen_nstr *nstr;
+  if (!(nstr = sen_nstr_open(str, str_len, encoding, flags))) {
+    return -1;
+  }
+  /* if the buffer size is short to store for the normalized string,
+     the required size is returned
+     (to inform the caller to cast me again). */
+  len = (int)nstr->norm_blen;
+  if (buf_size > len) {
+    memcpy(nstrbuf, nstr->norm, len + 1);
+  } else if (buf_size == len) {
+    /* NB: non-NULL-terminated */
+    memcpy(nstrbuf, nstr->norm, len);
+  }
+  sen_nstr_close(nstr);
+  return len;
+}
