@@ -25,10 +25,6 @@
 extern "C" {
 #endif
 
-#define SEN_STR_REMOVEBLANK 1
-#define SEN_STR_WITH_CTYPES 2
-#define SEN_STR_WITH_CHECKS 4
-
 enum {
   sen_str_null = 0,
   sen_str_alpha,
@@ -50,7 +46,7 @@ typedef struct {
   size_t length;
   int flags;
   sen_ctx *ctx;
-  /* sen_encoding encoding; */
+  sen_encoding encoding;
 } sen_nstr;
 
 typedef enum {
@@ -74,7 +70,6 @@ sen_nstr *sen_nstr_open(const char *str, size_t str_len, sen_encoding encoding,
 sen_nstr *sen_fakenstr_open(const char *str, size_t str_len, sen_encoding encoding, int flags);
 sen_rc sen_nstr_close(sen_nstr *nstr);
 
-size_t sen_str_charlen(const char *str, sen_encoding encoding);
 size_t sen_str_charlen_nonnull(const char *str, const char *end, sen_encoding encoding);
 size_t sen_str_len(const char *str, sen_encoding encoding, const char **last);
 sen_rc sen_str_fin(void);
