@@ -44,6 +44,7 @@ extern "C" {
 
 #define SEN_SNIP_NORMALIZE                      0x0001
 #define SEN_SNIP_COPY_TAG                       0x0002
+#define SEN_SNIP_SKIP_LEADING_SPACES            0x0004
 #define SEN_QUERY_SCAN_NORMALIZE                SEN_SNIP_NORMALIZE
 
 #define SEN_LEX_NGRAM_UNIT_SIZE                 2
@@ -130,6 +131,7 @@ typedef enum {
   sen_sel_partial,
   sen_sel_unsplit,
   sen_sel_near,
+  sen_sel_near2,
   sen_sel_similar,
   sen_sel_term_extract,
   sen_sel_prefix,
@@ -177,6 +179,7 @@ typedef struct _sen_snip_mapping sen_snip_mapping;
 typedef struct _sen_records_heap sen_records_heap;
 typedef struct _sen_vgram sen_vgram;
 typedef struct _sen_vgram_buf sen_vgram_buf;
+typedef struct _sen_sym_scan_hit sen_sym_scan_hit;
 typedef unsigned sen_id;
 
 #define SEN_ID_MAX 0x3fffffff
@@ -273,6 +276,12 @@ struct _sen_logger_info {
   void *func_arg;
 };
 
+struct _sen_sym_scan_hit {
+  sen_id id;
+  unsigned int offset;
+  unsigned int length;
+};
+
 typedef int (*query_term_callback)(const char *, unsigned int, void *);
 
 
@@ -446,6 +455,8 @@ sen_id sen_sym_common_prefix_search(sen_sym *sym, const void *key);
 int sen_sym_pocket_get(sen_sym *sym, sen_id id);
 sen_rc sen_sym_pocket_set(sen_sym *sym, sen_id id, unsigned int value);
 sen_id sen_sym_next(sen_sym *sym, sen_id id);
+int sen_sym_scan(sen_sym *sym, const char *str, unsigned int str_len,
+                 sen_sym_scan_hit *sh, unsigned int sh_size, const char **rest);
 
 /******** utility API ********/
 sen_snip *sen_snip_open(sen_encoding encoding, int flags, unsigned int width,
@@ -476,6 +487,15 @@ int sen_inv_entry_info(sen_inv *inv, sen_id key, unsigned *a, unsigned *pocket,
                        unsigned *size_in_chunk, unsigned *pos_in_chunk,
                        unsigned *size_in_buffer, unsigned *pos_in_buffer);
 
+/* flags for sen_str_normalize */
+#define SEN_STR_REMOVEBLANK 1
+#define SEN_STR_WITH_CTYPES 2
+#define SEN_STR_WITH_CHECKS 4
+int sen_str_normalize(const char *str, unsigned int str_len,
+                      sen_encoding encoding, int flags,
+                      char *nstrbuf, int buf_size);
+unsigned int sen_str_charlen(const char *str, sen_encoding encoding);
+
 /* misc */
 
 sen_rc sen_logger_info_set(const sen_logger_info *info);
